@@ -1,30 +1,29 @@
 """
-`Dataset`s and `LayerView`s form the core abstractions in mundipy.
+The `Dataset` forms the core abstraction in mundipy.
 
-`Dataset` comprises any source for vector data. Instantiating a
-`Dataset` declares its accessibility, but does not automatically
+A Dataset comprises any source for vector data. Instantiating a
+Dataset declares its accessibility, but does not automatically
 load features, as all features are lazily loaded.
 
-`LayerView` represents a collection of vector features, typically
-a subset from a `Dataset`. This makes queries like intersection
-and nearest much faster because only a subset of the `Dataset`
-must be loaded.
+Dataset implements the iterable interface, making it easy to iterate
+over features in a dataset.
 """
 
 from shapely.geometry.base import BaseGeometry
-from shapely.geometry import box, Point, Polygon, MultiPolygon
+from shapely.geometry import box, shape, Point, Polygon, MultiPolygon
 import shapely.wkt
 import shapely.wkb
-import geopandas as gpd
 from shapely.ops import transform
 from functools import lru_cache, partial
 from cached_property import cached_property_with_ttl
 import psycopg
+import fiona
 from psycopg_pool import ConnectionPool
 
 from mundipy.cache import (spatial_cache_footprint, pyproj_transform,
 	union_spatial_cache)
-from mundipy.geometry import from_dataframe, from_row_series, enrich_geom
+from mundipy.geometry import enrich_geom
+import mundipy.geometry as mgeom
 
 def elements_from_cursor(cur):
 	# get column names
@@ -68,23 +67,25 @@ class Dataset:
 		if isinstance(data, dict):
 			self._db_url = data['url']
 			self._db_table = data['table']
-
-			# 3 second timeout
-			self._pool = ConnectionPool(self._db_url, timeout=3.0)
 		elif isinstance(data, str):
 			self.filename = data
 		else:
 			raise TypeError('data for Dataset() is neither filename nor dict with PostgreSQL details')
 
+	@property
+	def _pool(self):
+		# 3 second timeout
+		return ConnectionPool(self._db_url, timeout=3.0, max_size=1)
+
 	@union_spatial_cache
-	def _load(self, geom, pcs='EPSG:4326'):
+	def _load(self, geom):
 		"""
 		Load part or the entire Dataset as a list of mundipy geometries.
 
 		Takes geom as a shapely.geometry, or None to load the
 		entire dataset.
 
-		Returns the dataset in PCS coordinates.
+		Returns the dataset in WGS84.
 		"""
 
 		if self._db_url is not None:
@@ -99,45 +100,33 @@ class Dataset:
 					# load entire geometry
 					query = "SELECT * FROM %s WHERE geometry && ST_GeomFromEWKT('SRID=4326;%s')" % (self._db_table, geom.wkt)
 
-				elements = elements_from_cursor(conn.execute(query))
-				return [geo.transform('EPSG:4326', pcs) for geo in elements]
+				return elements_from_cursor(conn.execute(query))
+
+		f = fiona.open(self.filename)
 
 		if geom is None:
-			gdf = gpd.read_file(self.filename)
+			items = f.values()
 		else:
-			gdf = gpd.read_file(self.filename, bbox=geom)
+			items = f.values(bbox=geom.bounds)
 
-		return [geo.transform('EPSG:4326', pcs) for geo in from_dataframe(gdf)]
+		return [ enrich_geom(shape(item['geometry']), dict(item['properties'])) for item in items ]
 
 	@lru_cache(maxsize=8)
-	def geometry_collection(self, pcs):
-		return self._load(None, pcs=pcs)
+	def geometry_collection(self):
+		return self._load(None)
 
 	"""Read into a Dataset at a specific geometry (WGS84)."""
-	def inside_bbox(self, bbox, pcs='EPSG:4326'):
+	def inside_bbox(self, bbox):
 		if not isinstance(bbox, tuple) or len(bbox) != 4:
 			raise TypeError('inside_bbox expected bbox to be a 4-tuple')
 
-		return self._load(box(*bbox), pcs=pcs)
-
-class LayerView:
-	"""
-	`LayerView` represents a collection of vector features in
-	a single dataset. It implements an `Iterable` interface,
-	allowing for one to loop through all features in the dataset,
-	or smart filtering without loading the entire dataset into
-	memory.
-	"""
-
-	def __init__(self, layer, pcs):
-		self.layer = layer
-		self.pcs = pcs
+		return self._load(box(*bbox))
 
 	def __iter__(self):
 		"""
 		Iterate through all items of the dataset.
 		"""
-		yield from self.layer.geometry_collection(self.pcs)
+		yield from self.geometry_collection()
 
 	def intersects(self, geom):
 		"""
@@ -151,17 +140,13 @@ class LayerView:
 		for feat in layer.intersects(Point(-37.0, 42.1)):
 			plot(feat)
 		"""
-		if not isinstance(self.layer, Dataset):
-			raise TypeError('intersects() on not Dataset undefined')
-		if not isinstance(geom, BaseGeometry):
-			raise TypeError('geom is not a shapely.geometry')
+		if not isinstance(geom, BaseGeometry) and not isinstance(geom, mgeom.BaseGeometry):
+			raise TypeError('geom is neither shapely.geometry nor mundipy.geometry')
 
-		# convert geom to EPSG:4326
-		to_wgs = pyproj_transform(self.pcs, 'EPSG:4326')
-		# buffer a little bit to prevent point
-		bbox = transform(to_wgs, geom.buffer(1e-8)).bounds
+		# buffer by an ~inch to prevent point
+		bbox = (geom.buffer(1e-3) if isinstance(geom, Point) or isinstance(geom, mgeom.Point) else geom).bounds
 
-		potentially_intersecting = self.layer.inside_bbox(bbox, self.pcs)
+		potentially_intersecting = self.inside_bbox(bbox)
 		return list(filter(lambda g: g.intersects(geom), potentially_intersecting))
 
 	def nearest(self, geom):
@@ -173,26 +158,66 @@ class LayerView:
 
 		`geom`: inherits from `shapely.geometry`
 		"""
-		if not isinstance(self.layer, Dataset):
-			raise TypeError('intersects() on not Dataset undefined')
-		if not isinstance(geom, BaseGeometry):
-			raise TypeError('geom is not a shapely.geometry')
-
-		to_wgs = pyproj_transform(self.pcs, 'EPSG:4326')
+		if not isinstance(geom, BaseGeometry) and not isinstance(geom, mgeom.BaseGeometry):
+			raise TypeError('geom is neither shapely.geometry nor mundipy.geometry')
 
 		# increasing look outside of this bbox for the nearest item
 		buffer_distances = [1e3, 1e4, 1e5, 1e6, 1e7, 1e8]
 		for buffer_size in buffer_distances:
 			# buffer geom.bbox
-			bbox = transform(to_wgs, geom.buffer(buffer_size)).bounds
+			bbox = geom.buffer(buffer_size).bounds
 
-			items = self.layer.inside_bbox(bbox, self.pcs)
+			items = self.inside_bbox(bbox)
 			if len(items) > 0:
 				return min(items, key=lambda geo: geom.distance(geo))
 
 		# fuck it, check the whole dataframe
-		items = self.layer.geometry_collection(self.pcs)
+		items = self.geometry_collection()
 		if len(items) > 0:
 			return min(items, key=lambda geo: geom.distance(geo))
 
 		return None
+
+	def _repr_svg_(self):
+		# Inspiration from https://github.com/shapely/shapely/blob/34136cd3104a197c3b47c97d44914197fb00879a/shapely/geometry/base.py#L263
+		# Some of the below code is licensed as shapely is
+
+		# Establish SVG canvas that will fit all the data + small space
+		total_bounds = [ obj.bounds for obj in iter(self) ]
+		total_bounds = ( min(map(lambda b: b[0], total_bounds)), min(map(lambda b: b[1], total_bounds)),
+						 max(map(lambda b: b[2], total_bounds)), max(map(lambda b: b[3], total_bounds)) )
+
+		xmin, ymin, xmax, ymax = total_bounds
+
+		# Expand bounds by a fraction of the data ranges
+		expand = 0.04  # or 4%, same as R plots
+		widest_part = max([xmax - xmin, ymax - ymin])
+		expand_amount = widest_part * expand
+		xmin -= expand_amount
+		ymin -= expand_amount
+		xmax += expand_amount
+		ymax += expand_amount
+
+		dx = xmax - xmin
+		dy = ymax - ymin
+		width = min([max([300.0, dx]), 500])
+		height = min([max([300.0, dy]), 500])
+		try:
+			scale_factor = max([dx, dy]) / max([width, height])
+		except ZeroDivisionError:
+			scale_factor = 1.0
+
+		view_box = f"{xmin} {ymin} {dx} {dy}"
+		transform = f"matrix(1,0,0,-1,0,{ymax + ymin})"
+
+		svg_content = ''.join(['<g transform="{0}">{1}</g>'.format(transform, child.svg(scale_factor)) for child in iter(self) ])
+
+		return (
+			'<svg xmlns="http://www.w3.org/2000/svg" '
+			'xmlns:xlink="http://www.w3.org/1999/xlink" '
+			'width="{1}" height="{2}" viewBox="{0}" '
+			'preserveAspectRatio="xMinYMin meet">'
+			'{3}</svg>'
+		).format(view_box, width, height, svg_content)
+
+
